import re
from collections import defaultdict
from typing import Callable, Dict, Optional, Tuple, Any, List, Set

from .ast import GrammarResolver, Node, Rule, Seq, Alt, Mult, Opt, Str, Rgx, \
    Look, NLook


class ParsingError(Exception):
    def __init__(self, msg: str, index: int, input: str):
        lines = input.split('\n')
        curr_len = 0
        for line_no, line in enumerate(lines):
            if curr_len + len(line) > index:
                pos = index - curr_len
                self.args = msg, (line_no + 1, pos), line
                return
            else:
                curr_len += len(line)
        self.args = msg, None, lines[-1]
        return

    def __str__(self):
        msg, loc, line = self.args
        if line[-1] != '\n':
            line += '\n'
        if loc:
            line_no, pos = loc
            context = f'at line {line_no}, pos {pos}:'
            pointer = (" " * (pos - 1)) + "/\\"
            return f'{msg} {context}\n{line}{pointer}'
        else:
            return f'{msg} at end of input:\n{line}'

    def prepend_msg(self, new_msg):
        msg, loc, line = self.args
        self.args = f'{new_msg}\n{msg}', loc, line


class Parser:
    grammar: Optional[Node]
    rules: Dict[str, Rule]
    actions: Dict[str, Callable]
    ignore_ws: bool

    def __init__(self, ignore_ws: bool = False):
        self.grammar = None
        self.rules = {}
        self.actions = {}
        self.ignore_ws = ignore_ws

    @staticmethod
    def from_grammar(grammar: str, *args, **kwargs):
        from .peg import peg_parser

        rules = peg_parser.parse_rule("grammar", grammar)

        parser = Parser(*args, **kwargs)
        parser.grammar = rules[0]
        parser.rules = {rule.name: rule for rule in rules}
        parser.link_rules()
        return parser

    def add_rule(self, pattern: str, action: Optional[Callable] = None):
        from .peg import peg_parser

        rule = peg_parser.parse_rule('rule', pattern)
        if self.grammar is None:
            self.grammar = rule

        name = rule.name
        self.rules[name] = rule
        if action:
            self.actions[name] = action

    def register_action(self, pattern: str):
        def wrap(action: Callable):
            self.add_rule(pattern, action)
            return action

        return wrap

    def link_rules(self):
        resolver = GrammarResolver()
        for rule in self.rules.values():
            rule.node = rule.node.visit(resolver, self.rules)

    def parse(self, input: str) -> Any:
        return self.parse_node(self.grammar, input)

    def parse_rule(self, name: str, input: str) -> Any:
        return self.parse_node(self.rules[name], input)

    def parse_node(self, node: Node, input: str) -> Any:
        run = ParserRun(self.actions, input, self.ignore_ws)
        res, end_index = node.visit(run, 0, [], set())
        if is_err(res):
            raise ParsingError(res.msg, end_index, input)
        if self.ignore_ws:
            end_index = run.skip_whitespace(end_index)
        if len(input) == end_index:
            return res
        raise ParsingError('Expected end of input', end_index, input)


PRes = Tuple[Any, int]


class Error:
    msg: str

    def __init__(self, msg):
        self.msg = msg

    def prepend_msg(self, msg):
        self.msg = msg + self.msg


def is_err(x):
    return type(x) == Error


class LeftRecursion:
    involved: Set[str]

    def __init__(self):
        self.involved = set()


def is_lr(x):
    return type(x) == LeftRecursion


class MemoEntry:
    res: Any
    idx: int

    def __init__(self, res: Any, idx: int):
        self.res = res
        self.idx = idx

    def __str__(self):
        return f"MemoEntry(res={self.res}, idx={self.idx})"

    def unwrap(self) -> Tuple[Any, int]:
        return self.res, self.idx


class ParserRun:
    actions: Dict[str, Callable]
    input: str
    ignore_ws: bool

    memotable: Dict[Tuple[str, int], MemoEntry]

    def __init__(self, actions, input: str, ignore_ws: bool):
        self.actions = actions
        self.input = input
        self.ignore_ws = ignore_ws

        self.memotable = defaultdict(lambda: None)

    def skip_whitespace(self, index: int) -> int:
        curr_index = index
        while True:
            char = self.input[curr_index:curr_index + 1]
            if len(char) > 0 and char.isspace():
                curr_index += 1
            else:
                return curr_index

    def apply_action(self, res, rule):
        if rule.name in self.actions and not is_err(res) and not is_lr(res):
            res = self.actions[rule.name](res)
        return res

    def grow_parse(self,
                   rule: Rule,
                   beg_idx: int,
                   stack: List[str],
                   involved: Set[str],
                   memo: MemoEntry) -> Tuple[Any, int]:
        while True:
            res, end_idx = rule.node.visit(
                self, beg_idx, stack, involved - {rule.name})
            if is_err(res) or end_idx <= memo.idx:
                break
            memo.res = self.apply_action(res, rule)
            memo.idx = end_idx
        return memo.unwrap()

    def get_involved_rules(self, stack: List[str], head: Rule) -> List[str]:
        involved = []
        for name in reversed(stack):
            involved.insert(0, name)
            if name == head.name:
                break
        return involved

    def visit_rule(self,
                   rule: Rule,
                   index: int,
                   stack: List[str],
                   involved: Set[Tuple[str, int]]) -> PRes:
        assert rule.node is not None, f'Rule {rule.name} does not have a body'

        memo = self.memotable[rule.name, index]
        if memo:
            if rule.name in involved:
                memo.res, memo.idx = rule.node.visit(
                    self, index, stack + [rule.name], involved - {rule.name})
                memo.res = self.apply_action(memo.res, rule)
                return memo.unwrap()
            elif is_lr(memo.res):
                path = self.get_involved_rules(stack, rule)
                str_path = "->".join(path)
                memo.res.involved |= set(path)
                msg = f'Infinite left recursion in path {str_path}'
                return Error(msg), index
            else:
                return memo.unwrap()
        else:
            lr = LeftRecursion()
            memo = MemoEntry(lr, index)
            self.memotable[rule.name, index] = memo
            memo.res, memo.idx = rule.node.visit(
                self, index, stack + [rule.name], involved)
            memo.res = self.apply_action(memo.res, rule)
            if lr.involved and not is_err(memo.res):
                return self.grow_parse(rule, index, stack, lr.involved, memo)
            else:
                return memo.unwrap()

    def visit_seq(self, seq: Seq, index: int, *args) -> PRes:
        res = []
        curr_index = index

        for node in seq.nodes:
            val, curr_index = node.visit(self, curr_index, *args)
            if is_err(val):
                return val, curr_index
            res.append(val)
        return res, curr_index

    def visit_alt(self, alt: Alt, index: int, *args) -> PRes:
        for node in alt.nodes:
            res, idx = node.visit(self, index, *args)
            if is_err(res):
                continue
            else:
                return res, idx
        return Error(f'No alternative matched in {alt}'), index

    def visit_mult(self, mult: Mult, index: int, *args) -> PRes:
        res = []
        curr_index = index

        while True:
            val, curr_index = mult.node.visit(self, curr_index, *args)
            if is_err(val):
                if len(res) < mult.min:
                    msg = f'{mult} matched fewer than {mult.min} time(s):\n'
                    val.prepend_msg(msg)
                    return val, curr_index
                else:
                    return res, curr_index
            else:
                res.append(val)

    def visit_opt(self, opt: Opt, index: int, *args) -> PRes:
        res, idx = opt.node.visit(self, index, *args)
        if is_err(res):
            return None, index
        else:
            return res, idx

    def visit_look(self, look: Look, index: int, *args) -> PRes:
        res, _ = look.node.visit(self, index, *args)
        return res, index

    def visit_nlook(self, nlook: NLook, index: int, *args) -> PRes:
        res, _ = nlook.node.visit(self, index, *args)
        if is_err(res):
            return None, index
        else:
            return Error(f'Did not expect {nlook.node}'), index

    def visit_str(self, string: Str, index: int, *args) -> PRes:
        if self.ignore_ws:
            index = self.skip_whitespace(index)
        str_len = len(string.string)
        if str_len <= len(self.input[index:]) and \
                string.string == self.input[index:index + str_len]:
            return string.string, index + str_len
        else:
            return Error(f'Expected `{string.string}`'), index

    def visit_rgx(self, regex: Rgx, index: int, *args) -> PRes:
        if self.ignore_ws:
            index = self.skip_whitespace(index)
        match = re.match(regex.pattern, self.input[index:])
        if match:
            string = match.group()
            return string, index + len(string)
        else:
            return Error(f'Could not match /{regex.pattern}/'), index



from dataclasses import dataclass
from typing import Union, List, Dict, Optional

Node = Union['Rule', 'Seq', 'Alt', 'Mult', 'Opt', 'Look', 'NLook', 'Str', 'Rgx']


@dataclass
class Rule:
    name: str
    node: Optional[Node] = None

    def __str__(self):
        return self.name

    def visit(self, visitor, *args, **kwargs):
        return visitor.visit_rule(self, *args, **kwargs)


@dataclass
class Seq:
    nodes: List[Node]

    def __init__(self, *nodes):
        self.nodes = list(nodes)

    def __str__(self):
        nodes = " ".join([str(node) for node in self.nodes])
        return f"( {nodes} )"

    def visit(self, visitor, *args, **kwargs):
        return visitor.visit_seq(self, *args, **kwargs)


@dataclass
class Alt:
    nodes: List[Node]

    def __init__(self, *nodes):
        self.nodes = list(nodes)

    def __str__(self):
        nodes = " | ".join([str(node) for node in self.nodes])
        return f"( {nodes} )"

    def visit(self, visitor, *args, **kwargs):
        return visitor.visit_alt(self, *args, **kwargs)


@dataclass
class Mult:
    min: int
    node: Node

    def __str__(self):
        if self.min == 0:
            symbol = "*"
        elif self.min == 1:
            symbol = "+"
        else:
            symbol = "{%d,}" % self.min
        return f"{self.node}{symbol}"

    def visit(self, visitor, *args, **kwargs):
        return visitor.visit_mult(self, *args, **kwargs)


@dataclass
class Opt:
    node: Node

    def __str__(self):
        return f"{self.node}?"

    def visit(self, visitor, *args, **kwargs):
        return visitor.visit_opt(self, *args, **kwargs)


@dataclass
class Look:
    node: Node

    def __str__(self):
        return f"&{self.node}"

    def visit(self, visitor, *args, **kwargs):
        return visitor.visit_look(self, *args, **kwargs)


@dataclass
class NLook:
    node: Node

    def __str__(self):
        return f"!{self.node}"

    def visit(self, visitor, *args, **kwargs):
        return visitor.visit_nlook(self, *args, **kwargs)


@dataclass
class Str:
    string: str

    def __str__(self):
        return self.string

    def visit(self, visitor, *args, **kwargs):
        return visitor.visit_str(self, *args, **kwargs)


@dataclass
class Rgx:
    pattern: str

    def __str__(self):
        return f"/{self.pattern}/"

    def visit(self, visitor, *args, **kwargs):
        return visitor.visit_rgx(self, *args, **kwargs)


class GrammarResolver:
    def visit_rule(self, rule: Rule, rules: Dict[str, Rule]) -> Rule:
        if rule.name in rules:
            return rules[rule.name]
        else:
            return rule

    def visit_seq(self, seq: Seq, rules) -> Seq:
        for i, node in enumerate(seq.nodes):
            seq.nodes[i] = node.visit(self, rules)
        return seq

    def visit_alt(self, alt: Alt, rules) -> Alt:
        for i, node in enumerate(alt.nodes):
            alt.nodes[i] = node.visit(self, rules)
        return alt

    def visit_mult(self, mult: Mult, rules) -> Mult:
        mult.node = mult.node.visit(self, rules)
        return mult

    def visit_opt(self, opt: Opt, rules) -> Opt:
        opt.node = opt.node.visit(self, rules)
        return opt

    def visit_look(self, look: Look, rules) -> Look:
        look.node = look.node.visit(self, rules)
        return look

    def visit_nlook(self, nlook: NLook, rules) -> NLook:
        nlook.node = nlook.node.visit(self, rules)
        return nlook

    def visit_str(self, string: Str, rules) -> Str:
        return string

    def visit_rgx(self, regex: Rgx, rules) -> Rgx:
        return regex



from typing import Any, Tuple, List

from .ast import Rule, Seq, Alt, Mult, Opt, Look, NLook, Str, Rgx
from .parser import Parser

rules = [
    Rule("grammar",
         Seq(Mult(1,
                  Seq(Rule("_"),
                      Rule("rule"),
                      Rule("_"),
                      Str(";"))),
             Rule("_"))),
    Rule("rule",
         Seq(Rule("name"),
             Rule("_"),
             Str("<-"),
             Rule("_"),
             Rule("alts"))),
    Rule("name",
         Rgx(r"[\w_-]+")),
    Rule("alts",
         Seq(Rule("seq"),
             Mult(0,
                  Seq(Rule("_"),
                      Str("|"),
                      Rule("_"),
                      Rule("seq"))))),
    Rule("seq",
         Seq(Rule("atom"),
             Mult(0,
                  Seq(Rule("_"),
                      Rule("atom"))))),
    Rule("atom",
         Alt(Rule("prefixed"),
             Rule("suffixed"),
             Rule("group"),
             Rule("string"),
             Rule("regex"),
             Rule("id"))),
    Rule("prefixed",
         Seq(Alt(Str("&"),
                 Str("!")),
             Rule("_"),
             Rule("atom"))),
    Rule("suffixed",
         Seq(Rule("atom"),
             Rule("_"),
             Alt(Str("+"),
                 Str("*"),
                 Str("?")))),
    Rule("group",
         Seq(Str("("),
             Rule("_"),
             Rule("alts"),
             Rule("_"),
             Str(")"))),
    Rule("string",
         Seq(Str('"'),
             Mult(0,
                  Alt(Str('\\"'),
                      Rule('dblslash'),
                      Rgx(r'[^\\"]+'))),
             Str('"'))),
    Rule("dblslash",
         Str(r"\\")),
    Rule("regex",
         Seq(Str("/"),
             Rgx(r'(\\/|[^/])*'),
             Str("/"))),
    Rule("id",
         Rgx(r"[\w_-]+")),
    Rule("_",
         Rgx("\\s*"))
]


def grammar_action(raw):
    rules, _ = raw
    return [rule for _, rule, _, _ in rules]


def rule_action(raw) -> Rule:
    name, _, _, _, body = raw
    return Rule(name, body)


def alts_action(raw) -> Alt:
    first = raw[0]
    rest = [alt for _, _, _, alt in raw[1]]

    if rest:
        alts = [first] + rest
        return Alt(*alts)
    else:
        return first


def seq_action(raw) -> Seq:
    first = raw[0]
    rest = [seq for _, seq in raw[1]]

    if rest:
        seqs = [first] + rest
        return Seq(*seqs)
    else:
        return first


def prefixed_action(raw):
    symbol, _, node = raw
    if symbol == "&":
        return Look(node)
    elif symbol == "!":
        return NLook(node)
    else:
        raise AssertionError(f"Unexpected suffix `{symbol}`")


def suffixed_action(raw):
    node, _, symbol = raw
    if symbol == "+":
        return Mult(1, node)
    elif symbol == "*":
        return Mult(0, node)
    elif symbol == "?":
        return Opt(node)
    else:
        raise AssertionError(f"Unexpected suffix `{symbol}`")


def string_action(raw: Tuple[str, List[str], str]):
    string = ''.join(raw[1])
    return Str(string)


def dblslash_action(raw: str):
    return "\\"


def middle(raw: Tuple[str, Any, str]) -> Any:
    return raw[1]


peg_parser = Parser()
peg_parser.rules = {rule.name: rule for rule in rules}
peg_parser.actions = {"grammar": grammar_action,
                      "rule": rule_action,
                      "alts": alts_action,
                      "seq": seq_action,
                      "prefixed": prefixed_action,
                      "suffixed": suffixed_action,
                      "group": lambda x: x[2],
                      "string": string_action,
                      "dblslash": dblslash_action,
                      "regex": lambda x: Rgx(middle(x)),
                      "id": lambda x: Rule(x)}
peg_parser.grammar = peg_parser.rules['rule']
peg_parser.link_rules()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="peg_leg",
    version="0.3.0",
    author="Gabriel Ionescu",
    author_email="gabe@erisian.tech",
    description="PEG parser generator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/formerly-a-trickster/peg-leg",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)